@@ -1,5 +1,9 @@
 %define mysql_version		@VERSION@
-%define release			0
+%ifarch i386
+%define release 0
+%else
+%define release 0.glibc23
+%endif
 %define mysqld_user		mysql
 %define server_suffix -standard
 
@@ -23,7 +27,7 @@ Packager:	Lenz Grimmer <build@mysql.com>
 Vendor:		MySQL AB
 Requires: fileutils sh-utils
 Provides:	msqlormysql MySQL-server mysql
-BuildPrereq: ncurses-devel
+BuildRequires: ncurses-devel
 Obsoletes:	mysql
 
 # Think about what you use here since the first step is to
@@ -77,9 +81,8 @@ The MySQL web site (http://www.mysql.com/) provides the latest
 news and information about the MySQL software. Also please see the
 documentation and the manual for more information.
 
-This package includes the MySQL server binary (statically linked,
-compiled with InnoDB support) as well as related utilities to run
-and administrate a MySQL server.
+This package includes the MySQL server binary (incl. InnoDB) as well
+as related utilities to run and administrate a MySQL server.
 
 If you want to access and work with the database, you have to install
 package "MySQL-client" as well!
@@ -101,6 +104,53 @@ This package contains the standard MySQL clients and administration tools.
 %description client -l pt_BR
 Este pacote contm os clientes padro para o MySQL.
 
+%package ndb-storage
+Release: %{release}
+Summary:	MySQL - ndbcluster storage engine
+Group:		Applications/Databases
+
+%description ndb-storage
+This package contains the ndbcluster storage engine. 
+It is necessary to have this package installed on all 
+computers that should store ndbcluster table data.
+Note that this storage engine can only be used in conjunction
+with the MySQL Max server.
+
+%{see_base}
+
+%package ndb-management
+Release: %{release}
+Summary:	MySQL - ndbcluster storage engine management
+Group:		Applications/Databases
+
+%description ndb-management
+This package contains ndbcluster storage engine management.
+It is necessary to have this package installed on at least 
+one computer in the cluster.
+
+%{see_base}
+
+%package ndb-tools
+Release: %{release}
+Summary:	MySQL - ndbcluster storage engine basic tools
+Group:		Applications/Databases
+
+%description ndb-tools
+This package contains ndbcluster storage engine basic tools.
+
+%{see_base}
+
+%package ndb-extra
+Release: %{release}
+Summary:	MySQL - ndbcluster storage engine extra tools
+Group:		Applications/Databases
+
+%description ndb-extra
+This package contains some extra ndbcluster storage engine tools for the advanced user.
+They should be used with caution.
+
+%{see_base}
+
 %package bench
 Release: %{release}
 Requires: %{name}-client perl-DBI perl
@@ -149,15 +199,23 @@ languages and applications need to dynamically load and use MySQL.
 
 %package Max
 Release: %{release}
-Summary: MySQL - server with Berkeley BD, RAID and UDF support
+Summary: MySQL - server with extended functionality
 Group: Applications/Databases
 Provides: mysql-Max
 Obsoletes: mysql-Max
 Requires: MySQL-server >= 4.0
 
 %description Max 
-Optional MySQL server binary that supports additional features like
-Berkeley DB, RAID and User Defined Functions (UDFs).
+Optional MySQL server binary that supports additional features like:
+
+ - Berkeley DB Storage Engine
+ - Ndbcluster Storage Engine interface
+ - Archive Storage Engine
+ - CSV Storage Engine
+ - Example Storage Engine
+ - MyISAM RAID
+ - User Defined Functions (UDFs).
+
 To activate this binary, just install this package in addition to
 the standard MySQL package.
 
@@ -189,9 +247,6 @@ client/server version.
 %setup -n mysql-%{mysql_version}
 
 %build
-# The all-static flag is to make the RPM work on different
-# distributions. This version tries to put shared mysqlclient libraries
-# in a separate package.
 
 BuildMySQL() {
 # The --enable-assembler simply does nothing on systems that does not
@@ -221,7 +276,7 @@ sh -c  "PATH=\"${MYSQL_BUILD_PATH:-$PATH}\" \
             --includedir=%{_includedir} \
             --mandir=%{_mandir} \
 	    --enable-thread-safe-client \
-	    --with-comment=\"Official MySQL RPM\";
+	    --with-readline ;
 	    # Add this for more debugging support
 	    # --with-debug
 	    # Add this for MyISAM RAID support:
@@ -268,10 +323,16 @@ then
 fi
 
 BuildMySQL "--enable-shared \
+		--without-openssl \
 		--with-berkeley-db \
 		--with-innodb \
+		--with-ndbcluster \
 		--with-raid \
+		--with-archive \
+		--with-csv-storage-engine \
+		--with-example-storage-engine \
 		--with-embedded-server \
+		--with-comment=\"MySQL Community Edition - Max (GPL)\" \
 		--with-server-suffix='-Max'"
 
 # Save everything for debug
@@ -281,6 +342,9 @@ BuildMySQL "--enable-shared \
 mv sql/mysqld sql/mysqld-max
 nm --numeric-sort sql/mysqld-max > sql/mysqld-max.sym
 
+# Install the ndb binaries
+(cd ndb; make install DESTDIR=$RBR)
+
 # Install embedded server library in the build root
 install -m 644 libmysqld/libmysqld.a $RBR%{_libdir}/mysql/
 
@@ -301,18 +365,21 @@ fi
 
 # Save manual to avoid rebuilding
 mv Docs/manual.ps Docs/manual.ps.save
-make distclean
+make clean
 mv Docs/manual.ps.save Docs/manual.ps
 
-# RPM:s destroys Makefile.in files, so we generate them here
-# aclocal; autoheader; aclocal; automake; autoconf
-# (cd innobase && aclocal && autoheader && aclocal && automake && autoconf)
-
-# Now build the statically linked 4.0 binary (which includes InnoDB)
+#
+# Only link statically on our i386 build host (which has a specially
+# patched static glibc installed) - ia64 and x86_64 run glibc-2.3 (unpatched)
+# so don't link statically there
+#
 BuildMySQL "--disable-shared \
+%ifarch i386
 		--with-mysqld-ldflags='-all-static' \
 		--with-client-ldflags='-all-static' \
 		$USE_OTHER_LIBC_DIR \
+%endif
+		--with-comment=\"MySQL Community Edition - Standard (GPL)\" \
 		--with-server-suffix='%{server_suffix}' \
 		--without-embedded-server \
 		--without-berkeley-db \
@@ -406,7 +473,7 @@ useradd -M -r -d $mysql_datadir -s /bin/bash -c "MySQL server" mysql 2> /dev/nul
 chown -R mysql $mysql_datadir
 
 # Initiate databases
-mysql_install_db -IN-RPM --user=mysql
+mysql_install_db --rpm --user=mysql
 
 # Change permissions again to fix any new files.
 chown -R mysql $mysql_datadir
@@ -421,6 +488,14 @@ chmod -R og-rw $mysql_datadir/mysql
 # Allow safe_mysqld to start mysqld and print a message before we exit
 sleep 2
 
+
+%post ndb-storage
+mysql_clusterdir=/var/lib/mysql-cluster
+
+# Create cluster directory if needed
+if test ! -d $mysql_clusterdir; then mkdir -m755 $mysql_clusterdir; fi
+
+
 %post Max
 # Restart mysqld, to use the new binary.
 echo "Restarting mysqld."
@@ -461,6 +536,7 @@ fi
 %doc Docs/manual.{html,ps,texi,txt}
 %doc Docs/manual_toc.html
 %doc support-files/my-*.cnf
+%doc support-files/ndb-*.ini
 
 %doc %attr(644, root, root) %{_infodir}/mysql.info*
 
@@ -484,12 +560,14 @@ fi
 %attr(755, root, root) %{_bindir}/myisamlog
 %attr(755, root, root) %{_bindir}/myisampack
 %attr(755, root, root) %{_bindir}/mysql_convert_table_format
+%attr(755, root, root) %{_bindir}/mysql_create_system_tables
 %attr(755, root, root) %{_bindir}/mysql_explain_log
 %attr(755, root, root) %{_bindir}/mysql_fix_extensions
 %attr(755, root, root) %{_bindir}/mysql_fix_privilege_tables
 %attr(755, root, root) %{_bindir}/mysql_install_db
 %attr(755, root, root) %{_bindir}/mysql_secure_installation
 %attr(755, root, root) %{_bindir}/mysql_setpermission
+%attr(755, root, root) %{_bindir}/mysql_tzinfo_to_sql
 %attr(755, root, root) %{_bindir}/mysql_zap
 %attr(755, root, root) %{_bindir}/mysqlbug
 %attr(755, root, root) %{_bindir}/mysqld_multi
@@ -540,6 +618,32 @@ fi
 %postun shared
 /sbin/ldconfig
 
+%files ndb-storage
+%defattr(-,root,root,0755)
+%attr(755, root, root) %{_sbindir}/ndbd
+
+%files ndb-management
+%defattr(-,root,root,0755)
+%attr(755, root, root) %{_sbindir}/ndb_mgmd
+%attr(755, root, root) %{_bindir}/ndb_mgm
+
+%files ndb-tools
+%defattr(-,root,root,0755)
+%attr(755, root, root) %{_bindir}/ndb_mgm
+%attr(755, root, root) %{_bindir}/ndb_restore
+%attr(755, root, root) %{_bindir}/ndb_waiter
+%attr(755, root, root) %{_bindir}/ndb_select_all
+%attr(755, root, root) %{_bindir}/ndb_select_count
+%attr(755, root, root) %{_bindir}/ndb_desc
+%attr(755, root, root) %{_bindir}/ndb_show_tables
+%attr(755, root, root) %{_bindir}/ndb_test_platform
+
+%files ndb-extra
+%defattr(-,root,root,0755)
+%attr(755, root, root) %{_bindir}/ndb_drop_index
+%attr(755, root, root) %{_bindir}/ndb_drop_table
+%attr(755, root, root) %{_bindir}/ndb_delete_all
+
 %files devel
 %defattr(-, root, root, 0755)
 %doc EXCEPTIONS-CLIENT
@@ -591,10 +695,31 @@ fi
 # itself - note that they must be ordered by date (important when
 # merging BK trees)
 %changelog 
+* Mon Feb 7 2005 Tomas Ulin <tomas@mysql.com>
+
+- enabled the "Ndbcluster" storage engine for the max binary
+- added extra make install in ndb subdir after Max build to get ndb binaries
+- added packages for ndbcluster storage engine
+
+* Fri Jan 14 2005 Lenz Grimmer <lenz@mysql.com>
+
+- replaced obsoleted "BuildPrereq" with "BuildRequires" instead
+
+* Thu Dec 31 2004 Lenz Grimmer <lenz@mysql.com>
+
+- enabled the "Archive" storage engine for the max binary
+- enabled the "CSV" storage engine for the max binary
+- enabled the "Example" storage engine for the max binary
+
 * Thu Aug 26 2004 Lenz Grimmer <lenz@mysql.com>
 
 - MySQL-Max now requires MySQL-server instead of MySQL (BUG 3860)
 
+* Fri Aug 20 2004 Lenz Grimmer <lenz@mysql.com>
+
+- do not link statically on IA64/AMD64 as these systems do not have
+  a patched glibc installed
+
 * Tue Aug 10 2004 Lenz Grimmer <lenz@mysql.com>
 
 - Added libmygcc.a to the devel subpackage (required to link applications
@@ -604,6 +729,21 @@ fi
 
 - Added EXCEPTIONS-CLIENT to the "devel" package
 
+* Thu Jul 29 2004 Lenz Grimmer <lenz@mysql.com>
+
+- disabled OpenSSL in the Max binaries again (the RPM packages were the
+  only exception to this anyway) (BUG 1043)
+
+* Wed Jun 30 2004 Lenz Grimmer <lenz@mysql.com>
+
+- fixed server postinstall (mysql_install_db was called with the wrong
+  parameter)
+
+* Thu Jun 24 2004 Lenz Grimmer <lenz@mysql.com>
+
+- added mysql_tzinfo_to_sql to the server subpackage
+- run "make clean" instead of "make distclean"
+
 * Mon Apr 05 2004 Lenz Grimmer <lenz@mysql.com>
 
 - added ncurses-devel to the build prerequisites (BUG 3377)
@@ -617,6 +757,10 @@ fi
 
 - added myisam_ftdump to the Server package
 
+* Tue Jan 13 2004 Lenz Grimmer <lenz@mysql.com>
+
+- link the mysql client against libreadline instead of libedit (BUG 2289)
+
 * Mon Dec 22 2003 Lenz Grimmer <lenz@mysql.com>
 
 - marked /etc/logrotate.d/mysql as a config file (BUG 2156)
@@ -629,6 +773,10 @@ fi
 
 - made testing for gcc3 a bit more robust
 
+* Fri Dec 05 2003 Lenz Grimmer <lenz@mysql.com>
+
+- added missing file mysql_create_system_tables to the server subpackage
+
 * Fri Nov 21 2003 Lenz Grimmer <lenz@mysql.com>
 
 - removed dependency on MySQL-client from the MySQL-devel subpackage
